import os
import os.path as op
import sys
import shutil
import argparse
import subprocess


# Directories/paths
this_dir = os.path.dirname(os.path.realpath(__file__))
template_origin_path = os.path.join(this_dir, 'xlwings_template.xltm')

if sys.platform.startswith('win'):
    win_template_path = op.join(os.getenv('APPDATA'), 'Microsoft', 'Templates', 'xlwings_template.xltm')
else:
    # Mac 2011 and 2016 use different directories
    from appscript import k, app
    from xlwings._xlmac import hfs_to_posix_path

    mac_template_dirs = set((op.realpath(op.join(op.expanduser("~"), 'Library', 'Application Support', 'Microsoft',
                                                 'Office', 'User Templates', 'My Templates')),
                             hfs_to_posix_path(app('Microsoft Excel').properties().get(k.templates_path))))

if sys.platform.startswith('win'):
    addin_path = os.path.join(os.getenv('APPDATA'), 'Microsoft', 'Excel', 'XLSTART', 'xlwings.xlam')


def addin_install(args):
    if not sys.platform.startswith('win'):
        print('Error: This command is only available on Windows right now.')
    else:
        try:
            shutil.copyfile(os.path.join(this_dir, 'xlwings.xlam'), addin_path)
            print('Successfully installed the xlwings add-in! Please restart Excel.')
        except IOError as e:
            if e.args[0] == 13:
                print('Error: Failed to install the add-in: If Excel is running, quit Excel and try again.')
            else:
                print(str(e))
        except Exception as e:
            print(str(e))


def addin_remove(args):
    if not sys.platform.startswith('win'):
        print('Error: This command is only available on Windows right now.')
    else:
        try:
            os.remove(addin_path)
            print('Successfully removed the xlwings add-in!')
        except WindowsError as e:
            if e.args[0] == 32:
                print('Error: Failed to remove the add-in: If Excel is running, quit Excel and try again.')
            elif e.args[0] == 2:
                print("Error: Could not remove the xlwings add-in. The add-in doesn't seem to be installed.")
            else:
                print(str(e))
        except Exception as e:
            print(str(e))


def addin_status(args):
    if not sys.platform.startswith('win'):
        print('Error: This command is only available on Windows right now.')
    else:
        if os.path.isfile(addin_path):
            print('The add-in is installed at {}'.format(addin_path))
            print('Use "xlwings addin remove" to uninstall it.')
        else:
            print('The add-in is not installed.')
            print('"xlwings addin install" will install it at: {}'.format(addin_path))


def template_open(args):
    if sys.platform.startswith('win'):
        subprocess.Popen('start {0}'.format(template_origin_path), shell=True)
    else:
        subprocess.Popen('open {0}'.format(template_origin_path), shell=True)


def template_install(args):
    if sys.platform.startswith('win'):
        try:
            shutil.copyfile(template_origin_path, win_template_path)
            print('Successfully installed the xlwings template')
        except Exception as e:
            print(str(e))
    else:
        for dir in mac_template_dirs:
            try:
                if os.path.isdir(dir):
                    path = op.realpath(op.join(dir, 'xlwings_template.xltm'))
                    shutil.copyfile(template_origin_path, path)
                    print('Successfully installed the xlwings template to {}'.format(path))
            except Exception as e:
                print('Error installing template to {}. {}'.format(path, str(e)))


def template_remove(args):
    if sys.platform.startswith('win'):
        try:
            os.remove(win_template_path)
            print('Successfully removed the xlwings template!')
        except WindowsError as e:
            print("Error: Could not remove the xlwings template. The template doesn't seem to be installed.")
        except Exception as e:
            print(str(e))
    else:
        for dir in mac_template_dirs:
            try:
                if os.path.isdir(dir):
                    path = op.realpath(op.join(dir, 'xlwings_template.xltm'))
                    os.remove(path)
                    print('Successfully removed the xlwings template from {}'.format(path))
            except OSError as e:
                print("Error: Could not remove the xlwings template. "
                      "The template doesn't seem to be installed at {}.".format(path))

            except Exception as e:
                print('Error removing template from {}. {}'.format(path, str(e)))


def template_status(args):
    if sys.platform.startswith('win'):
        if os.path.isfile(win_template_path):
            print('The template is installed at: {}'.format(win_template_path))
            print ('Use "xlwings template remove" to uninstall it.')
        else:
            print('The template can be installed at {}'.format(win_template_path))
            print('Use "xlwings template install" to install it or '
                  '"xlwings template open" to open it without installing.')
    else:
        is_installed = False
        can_be_installed = False
        for dir in mac_template_dirs:
            path = op.realpath(op.join(dir, 'xlwings_template.xltm'))
            if os.path.isfile(path):
                is_installed = True
                print('The template is installed at: {}'.format(path))
            else:
                if os.path.isdir(dir):
                    can_be_installed = True
                    print('The template can be installed at: {}'.format(dir))
        if can_be_installed:
            print('Use "xlwings template install" to install it or '
                  '"xlwings template open" to open it without installing.')
        if is_installed:
            print('Use "xlwings template remove" to uninstall it from all locations.')


def quickstart(args):
    project_name = args.project_name
    cwd = os.getcwd()

    # Project dir
    project_path = os.path.join(cwd, project_name)
    if not os.path.exists(project_path):
        os.makedirs(project_path)
    else:
        sys.exit('Error: Directory already exists.')

    # Python file
    with open(os.path.join(project_path, project_name + '.py'), 'w') as writer:
        writer.write('import xlwings as xw\n\n')

    # Excel file
    shutil.copyfile(os.path.join(this_dir, 'quickstart.xlsm'),
                    os.path.join(project_path, project_name + '.xlsm'))


def runpython_install(args):
    destination_dir = os.path.expanduser("~") + '/Library/Application Scripts/com.microsoft.Excel'
    if not os.path.exists(destination_dir):
        os.makedirs(destination_dir)
    shutil.copy(os.path.join(this_dir, 'xlwings.applescript'), destination_dir)
    print('Successfully installed RunPython for Mac Excel 2016!')


def main():
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest='command')
    subparsers.required = True

    # Add-in
    addin_parser = subparsers.add_parser('addin', help='xlwings Excel Add-in')
    addin_subparsers = addin_parser.add_subparsers(dest='subcommand')
    addin_subparsers.required = True
    
    addin_install_parser = addin_subparsers.add_parser('install')
    addin_install_parser.set_defaults(func=addin_install)

    addin_update_parser = addin_subparsers.add_parser('update')
    addin_update_parser.set_defaults(func=addin_install)

    addin_upgrade_parser = addin_subparsers.add_parser('upgrade')
    addin_upgrade_parser.set_defaults(func=addin_install)

    addin_remove_parser = addin_subparsers.add_parser('remove')
    addin_remove_parser.set_defaults(func=addin_remove)    

    addin_uninstall_parser = addin_subparsers.add_parser('uninstall')
    addin_uninstall_parser.set_defaults(func=addin_remove)

    addin_status_parser = addin_subparsers.add_parser('status')
    addin_status_parser.set_defaults(func=addin_status)

    # Template
    template_parser = subparsers.add_parser('template', help='xlwings Excel template')
    template_subparsers = template_parser.add_subparsers(dest='subcommand')
    template_subparsers.required = True

    template_open_parser = template_subparsers.add_parser('open')
    template_open_parser.set_defaults(func=template_open)

    template_install_parser = template_subparsers.add_parser('install')
    template_install_parser.set_defaults(func=template_install)

    template_update_parser = template_subparsers.add_parser('update')
    template_update_parser.set_defaults(func=template_install)

    template_remove_parser = template_subparsers.add_parser('remove')
    template_remove_parser.set_defaults(func=template_remove)

    template_uninstall_parser = template_subparsers.add_parser('uninstall')
    template_uninstall_parser.set_defaults(func=template_remove)

    template_status_parser = template_subparsers.add_parser('status')
    template_status_parser.set_defaults(func=template_status)

    # Quickstart
    quickstart_parser = subparsers.add_parser('quickstart', help='xlwings quickstart')
    quickstart_parser.add_argument("project_name")
    quickstart_parser.set_defaults(func=quickstart)

    # RunPython (only needed when installed with conda for Mac Excel 2016)
    if sys.platform.startswith('darwin'):
        runpython_parser = subparsers.add_parser('runpython', help='Run this if you installed xlwings via conda and are using Mac Excel 2016')
        runpython_subparser = runpython_parser.add_subparsers(dest='subcommand')
        runpython_subparser.required = True

        runpython_install_parser = runpython_subparser.add_parser('install')
        runpython_install_parser.set_defaults(func=runpython_install)

    args = parser.parse_args()
    args.func(args)

if __name__ == '__main__':
    main()



from .main import Range

expanders = {}

_empty = (None, '')


class Expander(object):

    def register(self, *aliases):
        for alias in aliases:
            expanders[alias] = self

    def expand(self, rng):
        """
        Expands a range

        Arguments
        ---------
        rng: Range
            The reference range

        Returns
        -------
        Range object: The expanded range

        """
        raise NotImplemented()

    def clear(self, rng, skip, vshape):
        """
        Clears out existing data corresponding to the expansion, in preparation for writing a value

        Arguments
        ---------
        rng: Range
            The reference range
        skip: tuple
            The number of rows, cols to skip when clearing (UDF formula caller range)
        vshape: tuple
            The number of rows, cols of the value which will be written subsequently
        """
        raise NotImplemented()


class TableExpander(Expander):

    def expand(self, rng):
        origin = rng(1, 1)
        if origin(2, 1).raw_value in _empty:
            bottom_left = origin
        elif origin(3, 1).raw_value in _empty:
            bottom_left = origin(2, 1)
        else:
            bottom_left = origin(2, 1).end('down')

        if origin(1, 2).raw_value in _empty:
            top_right = origin
        elif origin(1, 3).raw_value in _empty:
            top_right = origin(1, 2)
        else:
            top_right = origin(1, 2).end('right')

        return Range(top_right, bottom_left)

    def clear(self, rng, skip, vshape):

        # calculate how many rows of existing data are present
        xdata_origin = rng(1+skip[0], 1)
        if xdata_origin.raw_value in _empty:
            xdata_rows = 0
        elif xdata_origin(2, 1).raw_value in _empty:
            xdata_rows = 1
        else:
            xdata_rows = xdata_origin.end('down').row - xdata_origin.row + 1

        # calculate row to clear till
        clear_to_row = max(
            skip[0] + xdata_rows,  # clear out existing data
            1 + vshape[0]  # clear one space after last element
        )

        # calculate how many columns of existing data are present
        xdata_origin = rng(1, 1 + skip[1])
        if xdata_origin.raw_value in _empty:
            xdata_cols = 0
        elif xdata_origin(1, 2).raw_value in _empty:
            xdata_cols = 1
        else:
            xdata_cols = xdata_origin.end('right').column - xdata_origin.column + 1

        # calculate column to clear till
        clear_to_col = max(
            skip[1] + xdata_cols,  # clear out existing data
            1 + vshape[1]  # clear one space after last element
        )

        # We have now determined the shape of the rectangle to clear out, but there are
        # two top-left rectangular regions which require special treatment:
        # - skip: in the case of a UDF this is the shape occupied by the formula (or array formula)
        # - vshape: this is the shape of the value which will be written subsequently
        # In the first case we cannot clear the cells since that would cause the formula to be deleted,
        # in the second case there is no need to clear out those cells since their values will be overwritten.
        # The second case is not as indispensable like the first, however if we ignore it we get a flicker effect
        # as the value is written out.

        shapes = [skip, vshape]
        shapes.sort()

        # check if shape 0 is contained within shape 1
        if shapes[0][1] <= shapes[1][1]:
            shapes.pop(0)

        # check if either shape is degenerate (i.e. if at least one dimension is zero)
        for i in range(len(shapes)-1, -1, -1):
            if shapes[i][0] == 0 or shapes[i][1] == 0:
                shapes.pop(i)

        # add dummy corner at bottom-left of clear region, if needed
        if not shapes or shapes[-1][0] < clear_to_row:
            shapes.append((clear_to_row, 0))

        # clear the relevant cells, one row-block at a time, starting from the appropriate column
        #
        #    .....X
        #    ...XXX
        #    XXXXXX
        #
        prev_row = 1
        for s in shapes:
            if s[0] <= clear_to_row and s[1] < clear_to_col:
                Range(rng(prev_row, s[1] + 1), rng(s[0], clear_to_col)).clear_contents()
            prev_row = s[0] + 1

TableExpander().register('table')


class VerticalExpander(Expander):

    def expand(self, rng):
        if rng(2, 1).raw_value in _empty:
            return Range(rng(1, 1), rng(1, rng.shape[1]))
        elif rng(3, 1).raw_value in _empty:
            return Range(rng(1, 1), rng(2, rng.shape[1]))
        else:
            end_row = rng(2, 1).end('down').row - rng.row + 1
            return Range(rng(1, 1), rng(end_row, rng.shape[1]))

    def clear(self, rng, skip, vshape):

        # calculate how many rows of existing data are present
        xdata_origin = rng(1+skip[0], 1)
        if xdata_origin.raw_value in _empty:
            xdata_rows = 0
        elif xdata_origin(2, 1).raw_value in _empty:
            xdata_rows = 1
        else:
            xdata_rows = xdata_origin.end('down').row - xdata_origin.row + 1

        # calculate row to clear till
        clear_to_row = max(
            skip[0] + xdata_rows,       # clear out existing data
            1 + vshape[0]               # clear one space after last element
        )

        # calculate row to start clearing from
        clear_from_row = max(
            1 + skip[0],                # cannot clear before skip region
            1 + vshape[0]               # no point clearing rows where data will be written
        )

        if clear_from_row <= clear_to_row:
            ncols = len(rng.columns)
            Range(rng(clear_from_row, 1), rng(clear_to_row, ncols)).clear_contents()


VerticalExpander().register('vertical', 'down', 'd')


class HorizontalExpander(Expander):

    def expand(self, rng):
        if rng(1, 2).raw_value in _empty:
            return Range(rng(1, 1), rng(rng.shape[0], 1))
        elif rng(1, 3).raw_value in _empty:
            return Range(rng(1, 1), rng(rng.shape[0], 2))
        else:
            end_column = rng(1, 2).end('right').column - rng.column + 1
        return Range(rng(1, 1), rng(rng.shape[0], end_column))

    def clear(self, rng, skip, vshape):

        # calculate how many columns of existing data are present
        xdata_origin = rng(1, 1 + skip[1])
        if xdata_origin.raw_value in _empty:
            xdata_cols = 0
        elif xdata_origin(1, 2).raw_value in _empty:
            xdata_cols = 1
        else:
            xdata_cols = xdata_origin.end('right').column - xdata_origin.column + 1

        # calculate column to clear till
        clear_to_col = max(
            skip[1] + xdata_cols,   # clear out existing data
            1 + vshape[1]           # clear one space after last element
        )

        # calculate column to start clearing from
        clear_from_col = max(
            1 + skip[1],    # cannot clear before skip region
            1 + vshape[1]   # no point clearing columns where data will be written
        )

        if clear_from_col <= clear_to_col:
            nrows = len(rng.rows)
            Range(rng(1, clear_from_col), rng(nrows, clear_to_col)).clear_contents()


HorizontalExpander().register('horizontal', 'right', 'r')



import os
import sys
import re
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

# long_description: Take from README file
with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as f:
    readme = f.read()

# Version Number
with open(os.path.join(os.path.dirname(__file__), 'xlwings', '__init__.py')) as f:
    version = re.compile(r".*__version__ = '(.*?)'", re.S).match(f.read()).group(1)

# Dependencies
if sys.platform.startswith('win'):
    install_requires = ['comtypes']
    # This places dlls next to python.exe for standard setup and in the parent folder for virtualenv
    data_files = [('', ['xlwings32.dll', 'xlwings64.dll'])]
elif sys.platform.startswith('darwin'):
    install_requires = ['psutil >= 2.0.0', 'appscript >= 1.0.1']
    data_files = [(os.path.expanduser("~") + '/Library/Application Scripts/com.microsoft.Excel', ['xlwings/xlwings.applescript'])]
else:
    on_rtd = os.environ.get('READTHEDOCS', None) == 'True'
    if on_rtd:
        data_files = []
        install_requires = []
    else:
        raise OSError("currently only Windows and OSX are supported.")

# This shouldn't be necessary anymore as we dropped official support for < 2.7 and < 3.3
if (sys.version_info[0] == 2 and sys.version_info[:2] < (2, 7)) or (sys.version_info[0] == 3 and sys.version_info[:2] < (3, 2)):
    install_requires = install_requires + ['argparse']

setup(
    name='xlwings',
    version=version,
    url='http://xlwings.org',
    license='BSD 3-clause',
    author='Zoomer Analytics LLC',
    author_email='felix.zumstein@zoomeranalytics.com',
    description='Make Excel fly: Interact with Excel from Python and vice versa.',
    long_description=readme,
    data_files=data_files,
    packages=['xlwings', 'xlwings.tests', 'xlwings.conversion'],
    package_data={'xlwings': ['xlwings.bas', 'tests/*.xlsx', 'tests/*.xlsm', 'tests/*.png', 'xlwings_template.xltm',
                              'quickstart.xlsm', 'xlwings.xlam', 'xlwings.applescript']},
    keywords=['xls', 'excel', 'spreadsheet', 'workbook', 'vba', 'macro'],
    install_requires=install_requires,
    entry_points={'console_scripts': ['xlwings=xlwings.command_line:main'],},
    classifiers=[
        'Development Status :: 4 - Beta',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: MacOS :: MacOS X',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Topic :: Office/Business :: Financial :: Spreadsheet',
        'License :: OSI Approved :: BSD License'],
    platforms=['Windows', 'Mac OS X'],
)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals
import unittest

from xlwings.tests.common import TestBase


class TestNames(TestBase):
    def test_get_names_index(self):
        self.wb1.sheets[0].range('B2:D10').name = 'test1'
        self.wb1.sheets[0].range('A1').name = 'test2'
        self.assertEqual(self.wb1.names(1).name, 'test1')
        self.assertEqual(self.wb1.names[1].name, 'test2')

    def test_names_contain(self):
        self.wb1.sheets[0].range('B2:D10').name = 'test1'
        self.assertTrue('test1' in self.wb1.names)

    def test_len(self):
        self.wb1.sheets[0].range('B2:D10').name = 'test1'
        self.wb1.sheets[0].range('A1').name = 'test2'
        self.assertEqual(len(self.wb1.names), 2)

    def test_count(self):
        self.assertEqual(len(self.wb1.names), self.wb1.names.count)

    def test_names_iter(self):
        self.wb1.sheets[0].range('B2:D10').name = 'test1'
        self.wb1.sheets[0].range('A1').name = 'test2'
        for ix, n in enumerate(self.wb1.names):
            if ix == 0:
                self.assertEqual(n.name, 'test1')
            if ix == 1:
                self.assertEqual(n.name, 'test2')

    def test_get_inexisting_name(self):
        self.assertIsNone(self.wb1.sheets[0].range('A1').name)

    def test_get_set_named_range(self):
        self.wb1.sheets[0].range('A100').name = 'test1'
        self.assertEqual(self.wb1.sheets[0].range('A100').name.name, 'test1')

        self.wb1.sheets[0].range('A200:B204').name = 'test2'
        self.assertEqual(self.wb1.sheets[0].range('A200:B204').name.name, 'test2')

    def test_delete_named_item1(self):
        self.wb1.sheets[0].range('B10:C11').name = 'to_be_deleted'
        self.assertEqual(self.wb1.sheets[0].range('to_be_deleted').name.name, 'to_be_deleted')

        del self.wb1.names['to_be_deleted']
        self.assertIsNone(self.wb1.sheets[0].range('B10:C11').name)

    def test_delete_named_item2(self):
        self.wb1.sheets[0].range('B10:C11').name = 'to_be_deleted'
        self.assertEqual(self.wb1.sheets[0].range('to_be_deleted').name.name, 'to_be_deleted')

        self.wb1.names['to_be_deleted'].delete()
        self.assertIsNone(self.wb1.sheets[0].range('B10:C11').name)

    def test_delete_named_item3(self):
        self.wb1.sheets[0].range('B10:C11').name = 'to_be_deleted'
        self.assertEqual(self.wb1.sheets[0].range('to_be_deleted').name.name, 'to_be_deleted')

        self.wb1.sheets[0].range('to_be_deleted').name.delete()
        self.assertIsNone(self.wb1.sheets[0].range('B10:C11').name)

    def test_names_collection(self):
        self.wb1.sheets[0].range('A1').name = 'name1'
        self.wb1.sheets[0].range('A2').name = 'name2'
        self.assertTrue('name1' in self.wb1.names and 'name2' in self.wb1.names)

        self.wb1.sheets[0].range('A3').name = 'name3'
        self.assertTrue('name1' in self.wb1.names and 'name2' in self.wb1.names and
                    'name3' in self.wb1.names)

    def test_sheet_scope(self):
        self.wb2.sheets[0].range('B2:C3').name = 'Sheet1!sheet_scope1'
        self.wb2.sheets[0].range('sheet_scope1').value = [[1., 2.], [3., 4.]]
        self.assertEqual(self.wb2.sheets[0].range('B2:C3').value, [[1., 2.], [3., 4.]])
        with self.assertRaises(Exception):
            values = self.wb2.sheets[1].range('sheet_scope1').value

    def test_workbook_scope(self):
        self.wb1.sheets[0].range('A1').name = 'test1'
        self.wb1.sheets[0].range('test1').value = 123.
        self.assertEqual(self.wb1.names['test1'].refers_to_range.value, 123.)

    def test_contains_name(self):
        self.wb1.sheets[0].range('A1').name = 'test1'
        self.assertTrue(self.wb1.names.contains('test1'))
        self.assertFalse(self.wb1.names.contains('test2'))

    def test_names_add(self):
        self.wb1.names.add('test1', '=Sheet1!$A$1:$B$3')
        self.assertEqual(self.wb1.sheets[0].range('A1:B3').name.name, 'test1')

    def test_refers_to_range(self):
        self.wb1.sheets[0].range('B2:D10').name = 'test1'
        self.assertEqual(self.wb1.sheets[0].range('B2:D10').address, self.wb1.sheets[0].range('B2:D10').name.refers_to_range.address)

if __name__ == '__main__':
    unittest.main()
